library jinja.loaders;

import 'dart:io';

import 'package:path/path.dart' as _path;

import 'env.dart';

typedef Template TemplateLoader(String path);

abstract class Loader {
  String getSource(String path) {
    throw Exception('Template not found: $path');
  }

  bool get hasSourceAccess => true;

  List<String> listTemplates() {
    throw Exception('This loader cannot iterate over all templates');
  }

  void load(Env env) {
    for (var path in listTemplates())
      env.templates[path] = Template(getSource(path), path);
  }

  @override
  String toString() => 'Loader()';
}

class FileSystemLoader extends Loader {
  FileSystemLoader(String path) : dir = Directory(path) {
    if (!dir.existsSync()) throw Exception('Directory not exist: $path');
  }

  final Directory dir;

  @override
  String getSource(String path) =>
      File(_path.join(dir.path, path)).readAsStringSync();

  @override
  List<String> listTemplates() => dir
      .listSync(recursive: true)
      .map<String>((entity) => _path.relative(entity.path, from: dir.path))
      .toList();

  @override
  void load(Env env) {
    super.load(env);
    if (env.debug) {
      dir
          .watch(recursive: true)
          .where((event) => event.type == FileSystemEvent.modify)
          .listen((event) {
        var path = _path.relative(event.path, from: dir.path);
        env.templates[path] = Template(getSource(path), path);
        print('Template reloaded: $path');
      });
    }
  }

  @override
  String toString() => 'FileSystemLoader($dir)';
}

class DictLoader extends Loader {
  DictLoader(this.dict);

  final Map<String, String> dict;

  final bool hasSourceAccess = false;

  @override
  List<String> listTemplates() => dict.keys.toList();

  @override
  String getSource(String path) {
    if (dict.containsKey(path)) return dict[path];
    throw Exception('Template not found: $path');
  }

  @override
  String toString() => 'DictLoader($dict)';
}
