import 'package:string_scanner/string_scanner.dart';

import 'env.dart';
import 'nodes.dart';
import 'utils.dart';

class _Params {
  const _Params(this.args, this.kwargs);

  final List<Expression> args;
  final Map<String, Expression> kwargs;
}

class Parser {
  final _stmtOpenReg =
      RegExp(RegExp.escape(Env.instance.stmtOpen) + _spaceReg.pattern);
  final _stmtCloseReg =
      RegExp(_spaceReg.pattern + RegExp.escape(Env.instance.stmtClose));
  final _varOpenReg =
      RegExp(RegExp.escape(Env.instance.varOpen) + _spaceReg.pattern);
  final _varCloseReg =
      RegExp(_spaceReg.pattern + RegExp.escape(Env.instance.varClose));
  final _commentOpenReg =
      RegExp(RegExp.escape(Env.instance.commentOpen) + _spaceReg.pattern);
  final _commentCloseReg =
      RegExp(_spaceReg.pattern + RegExp.escape(Env.instance.commentClose));

  Parser(String source, {dynamic uri})
      : scanner = SpanScanner(source, sourceUrl: uri);

  final SpanScanner scanner;

  final _buffer = StringBuffer();
  final _nodes = <Node>[];

  Node parse() {
    while (!scanner.isDone) {
      if (scanner.matches(RegExp([
        _stmtOpenReg.pattern,
        _varOpenReg.pattern,
        _commentOpenReg.pattern
      ].join('|')))) {
        _flush();
        if (scanner.scan(_stmtOpenReg))
          _parseStmt(scanner);
        else if (scanner.scan(_varOpenReg))
          _parseVar(scanner);
        else if (scanner.scan(_commentOpenReg))
          _skipComment(scanner);
        else
          scanner.error('Can not match');
      } else
        _buffer.writeCharCode(scanner.readChar());
    }
    _flush();
    if (_nodes.isEmpty) return Literal<Null>(null);
    Node first = _nodes.first;
    if (first is Tag && first.tag == 'extends') {
      _nodes.removeAt(0);
      if (!_nodes.every((node) => node is! Extend))
        throw Exception('child template only contains one extends statement');
      return Extend(
          first.arg as Expression,
          _nodes.whereType<Block>().toList().asMap().map<String, Block>(
              (_, block) => MapEntry<String, Block>(block.name, block)));
    }
    return Output.nodeOrOutput(_nodes);
  }

  void _parseStmt(SpanScanner scanner) {
    scanner.expect(_identifierReg);
    String tag = scanner.lastMatch[1];
    scanner.scan(_spacePlusReg);
    Node node;
    if (tag.startsWith('end')) {
      tag = tag.substring(3);
      switch (tag) {
        case 'block':
          node = _parseBlockStmt();
          break;
        case 'for':
          node = _parseForStmt();
          break;
        case 'if':
          node = _parseIfStmt();
          break;
        default:
          throw FormatException('Unimplemented tag: $tag');
      }
    } else
      switch (tag) {
        case 'extends':
          if (_nodes.isNotEmpty && _nodes.first is Extend)
            scanner.error('extends can be only one and first');
          Expression parent = _parsePrimaryExpr(scanner);
          node = Tag(tag, parent);
          break;
        case 'block':
          scanner.expect(_identifierReg, name: 'block name');
          String name = scanner.lastMatch[1];
          node = Tag(tag, name);
          break;
        case 'for':
          scanner.expect(_identifierReg);
          String target = scanner.lastMatch[1];
          if (target == 'in') throw FormatException();
          scanner.expect(_inReg);
          Expression iter = _parsePrimaryExpr(scanner);
          if (iter == null)
            throw FormatException('Expected for iterable expression');
          if (scanner.scan(_ifReg)) {
            Expression test = _parseExpr(scanner);
            if (test == null)
              throw FormatException('Expected for test expression');
            if (test is! Test) test = Test.name('defined', test);
            node = Tag(tag, target, iter, test);
          } else
            node = Tag(tag, target, iter);
          break;
        case 'if':
        case 'elif':
          Expression test = _parseExpr(scanner);
          if (test is! Test) test = Test.name('defined', test);
          node = Tag(tag, test);
          break;
        case 'else':
          node = Tag(tag);
          break;
        default:
          throw FormatException('Unimplemented tag: $tag');
      }
    scanner.expect(_stmtCloseReg);
    _nodes.add(node);
  }

  Node _parseBlockStmt() {
    String name;
    Node body;
    List<Node> nodes = <Node>[];
    while (_nodes.isNotEmpty) {
      Node node = _nodes.removeLast();
      if (node is Tag && node.tag == 'block') {
        name = node.arg as String;
        body = Output.nodeOrOutput(nodes.reversed.toList());
        nodes.clear();
        break;
      } else
        nodes.add(node);
    }
    return Block(name, body);
  }

  Node _parseForStmt() {
    String target;
    Node body, orElse;
    Expression iter;
    Test test;
    List<Node> nodes = <Node>[];
    while (_nodes.isNotEmpty) {
      Node node = _nodes.removeLast();
      if (node is Tag) if (node.tag == 'else') {
        orElse = Output.nodeOrOutput(nodes.reversed.toList());
        nodes.clear();
      } else if (node.tag == 'for') {
        target = node.arg as String;
        iter = node.arg2 as Expression;
        body = Output.nodeOrOutput(nodes.reversed.toList());
        test = node.arg3 as Test;
        nodes.clear();
        break;
      } else
        throw FormatException(node.toString());
      else
        nodes.add(node);
    }
    return For(target, iter, body, test: test, orElse: orElse);
  }

  Node _parseIfStmt() {
    Test test;
    Node body, orElse;
    List<Node> nodes = <Node>[];
    while (_nodes.isNotEmpty) {
      Node node = _nodes.removeLast();
      if (node is Tag) if (node.tag == 'else') {
        orElse = Output.nodeOrOutput(nodes.reversed.toList());
        nodes.clear();
      } else if (node.tag == 'elif') {
        test = node.arg as Test;
        body = Output.nodeOrOutput(nodes.reversed.toList());
        nodes.clear();
        orElse = If(test, body, orElse);
      } else if (node.tag == 'if') {
        test = node.arg as Test;
        body = Output.nodeOrOutput(nodes.reversed.toList());
        nodes.clear();
        break;
      } else
        throw FormatException(node.toString());
      else
        nodes.add(node);
    }
    return If(test, body, orElse);
  }

  void _parseVar(SpanScanner scanner) {
    Expression expr = _parseExpr(scanner);
    if (expr == null) throw FormatException('Expected expression');
    _nodes.add(expr);
    scanner.expect(_varCloseReg, name: 'double curly');
  }

  void _skipComment(SpanScanner scanner) {
    while (!scanner.scan(_commentCloseReg)) scanner.position++;
  }

  void _flush() {
    if (_buffer.isNotEmpty) {
      String body = _buffer.toString();
      _buffer.clear();
      _nodes.add(Literal<String>(body));
    }
  }

  Expression _parseExpr(SpanScanner scanner) {
    Expression expr;
    if (!isNull(expr = _parseLogicalOrExpr(scanner))) {
      while (true) {
        if (scanner.scan(_pipeReg)) {
          scanner.expect(_identifierReg);
          String name = scanner.lastMatch[1];
          if (scanner.scan(_lParenReg)) {
            _Params params = _parseParams(scanner);
            scanner.expect(_rParenReg, name: 'Expected paren close');
            expr = Filter.name(name, expr,
                args: params.args, kwargs: params.kwargs);
          } else
            expr = Filter.name(name, expr);
        } else
          break;
      }
      return expr;
    }
    return expr;
  }

  List<Expression> _parseExprList(SpanScanner scanner) {
    Expression expr;
    List<Expression> list = <Expression>[];
    do {
      expr = _parseExpr(scanner);
      if (expr != null) list.add(expr);
    } while (scanner.scan(_commaReg));
    if (list.isNotEmpty) return list;
    return null;
  }

  // Expression _parseConditionalExpr(SpanScanner scanner) {
  //   Expression expr;
  //   if (!isNull(expr = parseLogicalOrExpr(scanner))) {
  //     if (scanner.scan(_ifReg)) {
  //       Expression condition;
  //       if (isNull(condition = parseExpr(scanner)))
  //         scanner.error('Expected expression');
  //       if (scanner.scan(_isReg)) {
  //         bool isNot = scanner.scan(_notReg);
  //         scanner.expect(_identifierReg);
  //         String name = scanner.lastMatch[1];
  //         if (scanner.scan(_lParenReg)) {
  //           List<Expression> args = parseExprList(scanner);
  //           if (isNull(expr)) scanner.error('Expected expression');
  //           scanner.expect(_rParenReg);
  //           Expression test = Test.name(name, expr, args: args);
  //           condition = isNot ? Not(test) : test;
  //         }
  //         scanner.scan(_spaceReg);
  //         if (!scanner.matches(_elseReg)) {
  //           Expression arg = parsePrimaryExpr(scanner);
  //           condition = isNull(arg)
  //               ? Test.name(name, expr)
  //               : Test.name(name, expr, args: [arg]);
  //         }
  //       } else if (scanner.scan(_notInReg)) {
  //         bool isNot = scanner.lastMatch[2] == 'not';
  //         Expression seq = parsePrimaryExpr(scanner);
  //         if (isNull(seq)) scanner.error('Expected primary expression');
  //         Expression test = Test.name('in', expr, args: [seq]);
  //         condition = isNot ? Not(test) : test;
  //       }
  //       scanner.expect(_elseReg);
  //       Expression orElse;
  //       if (isNull(orElse = parseExpr(scanner)))
  //         scanner.error('Expected expression');
  //       return Conditional(expr, condition, orElse);
  //     }
  //   }
  //   return expr;
  // }

  Expression _parseLogicalOrExpr(SpanScanner scanner) {
    Expression left;
    if (!isNull(left = _parseLogicalAndExpr(scanner))) {
      while (scanner.scan(_logicalOrReg)) {
        Expression right;
        if (isNull(right = _parseLogicalAndExpr(scanner)))
          scanner.error('Expected expression');
        left = Or(left, right);
      }
      return left;
    }
    return left;
  }

  Expression _parseLogicalAndExpr(SpanScanner scanner) {
    Expression left;
    if (!isNull(left = _parseEqualityExpr(scanner))) {
      Expression right;
      while (scanner.scan(_logicalAndReg)) {
        if (isNull(right = _parseEqualityExpr(scanner)))
          scanner.error('Expected equality expression');
        left = And(left, right);
      }
      return left;
    }
    return left;
  }

  Expression _parseEqualityExpr(SpanScanner scanner) {
    Expression left;
    if (!isNull(left = _parseRelationalExpr(scanner))) {
      if (scanner.scan(_equalityReg)) {
        String op = scanner.lastMatch[1];
        Expression right;
        if (isNull(right = _parseRelationalExpr(scanner)))
          scanner.error('Expected relational expression');
        switch (op) {
          case '==':
            return Equal(left, right);
          case '!=':
            return NotEqual(left, right);
        }
      }
    }
    return left;
  }

  Expression _parseRelationalExpr(SpanScanner scanner) {
    Expression left;
    if (!isNull(left = _parseAdditiveExpr(scanner))) {
      if (scanner.scan(_relationalReg)) {
        Expression right;
        String op = scanner.lastMatch[1];
        if (isNull(right = _parseAdditiveExpr(scanner)))
          scanner.error('Expected additive expression');
        switch (op) {
          case '>=':
            return GreaterOrEqual(left, right);
          case '>':
            return Greater(left, right);
          case '<':
            return Less(left, right);
          case '<=':
            return LessOrEqual(left, right);
        }
      }
    }
    return left;
  }

  Expression _parseAdditiveExpr(SpanScanner scanner) {
    Expression left;
    if (!isNull(left = _parseMultiplicativeExpr(scanner))) {
      Expression right;
      String op;
      while (scanner.scan(_additiveReg)) {
        op = scanner.lastMatch[1];
        if (isNull(right = _parseMultiplicativeExpr(scanner)))
          scanner.error('Expected multiplicative expression');
        switch (op) {
          case '+':
            left = Add(left, right);
            break;
          case '-':
            left = Sub(left, right);
            break;
        }
      }
    }
    return left;
  }

  Expression _parseMultiplicativeExpr(SpanScanner scanner) {
    Expression left;
    if (!isNull(left = _parseUnaryExpression(scanner))) {
      Expression right;
      String op;
      while (scanner.scan(_multiplicativeReg)) {
        op = scanner.lastMatch[1];
        if (isNull(right = _parseUnaryExpression(scanner)))
          scanner.error('Expected unary expression');
        switch (op) {
          case '*':
            left = Mul(left, right);
            break;
          case '/':
            left = Div(left, right);
            break;
          case '%':
            left = Mod(left, right);
            break;
          case '//':
            left = FloorDiv(left, right);
            break;
          case '**':
            left = Pow(left, right);
            break;
        }
      }
      return left;
    }
    return left;
  }

  Expression _parseUnaryExpression(SpanScanner scanner) {
    Expression expr;
    if (scanner.scan(_unaryReg) || scanner.scan(_notReg)) {
      String op = scanner.lastMatch[1];
      if (isNull(expr = _parsePostfixExpr(scanner)))
        scanner.error('Expected unary expression');
      switch (op) {
        case 'not':
          return Not(expr);
        case '-':
          return Neg(expr);
        case '+':
          return Pos(expr);
      }
    }
    if (!isNull(expr = _parsePostfixExpr(scanner))) return expr;
    return null;
  }

  Expression _parsePostfixExpr(SpanScanner scanner) {
    Expression expr;
    if (!isNull(expr = _parsePrimaryExpr(scanner))) {
      while (true) {
        if (scanner.scan(_lBracketReg)) {
          Expression key = _parseExpr(scanner);
          if (isNull(key)) scanner.error('Expected expression');
          if (scanner.scan(_commaReg)) {
            if (key is Literal<num>) {
              Expression endKey = _parseExpr(scanner);
              expr = Filter.name('sublist', expr, args: [key, endKey]);
            } else
              scanner.error('slice args must be num');
          } else {
            expr = Key(key, expr);
          }
          scanner.expect(_rBracketReg, name: ']');
        } else if (scanner.scan(_fieldReg))
          expr = Field(scanner.lastMatch[1], expr);
        else if (scanner.scan(_lParenReg)) {
          _Params params = _parseParams(scanner);
          scanner.expect(_rParenReg, name: ')');
          expr = Call(expr, args: params.args, kwargs: params.kwargs);
        } else
          break;
      }
      return expr;
    }
    return expr;
  }

  _Params _parseParams(SpanScanner scanner) {
    Expression expr;
    List<Expression> args = <Expression>[];
    Map<String, Expression> kwargs = <String, Expression>{};
    bool inKeys = false;
    do {
      if (scanner.scan(_identifierReg)) {
        String key = scanner.lastMatch[1];
        if (scanner.scan(_equalSignReg)) {
          inKeys = true;
          if (isNull(expr = _parseExpr(scanner)))
            scanner.error('Expected expression');
          kwargs[key] = expr;
        }
      } else if (inKeys)
        scanner.error('Expected key=value entry');
      else {
        expr = _parseExpr(scanner);
        if (!isNull(expr)) args.add(expr);
      }
    } while (scanner.scan(_commaReg));
    return _Params(args, kwargs);
  }

  Expression _parsePrimaryExpr(SpanScanner scanner) {
    Expression expr;
    if (!isNull(expr = _parseCompoundLiteral(scanner))) return expr;
    if (!isNull(expr = _parseLiteral(scanner))) return expr;
    if (scanner.scan(_identifierReg)) return Variable(scanner.lastMatch[1]);
    return null;
  }

  Literal _parseLiteral(SpanScanner scanner) {
    if (scanner.scan('none')) return const Literal<Null>(null);
    if (scanner.scan('false')) return const Literal<bool>(false);
    if (scanner.scan('true')) return const Literal<bool>(true);
    if (scanner.scan(_digitReg)) {
      String rawNum = scanner.lastMatch[1];
      if (scanner.scan(_fractionalReg)) rawNum += scanner.lastMatch[1];
      return Literal<num>(num.tryParse(rawNum));
    }
    if (scanner.scan(_stringStartReg)) {
      String body;
      switch (scanner.lastMatch[1]) {
        case '"':
          scanner.expect(_stringContentDQReg,
              name: 'Expected optional string content and double quote');
          body = scanner.lastMatch[1];
          break;
        case "'":
          scanner.expect(_stringContentSQReg,
              name: 'Expected optional string content and single quote');
          body = scanner.lastMatch[1];
          break;
      }
      return Literal<String>(body);
    }
    return null;
  }

  Expression _parseCompoundLiteral(SpanScanner scanner) {
    Expression expr;
    if (!isNull(expr = _parseListLiteral(scanner))) return expr;
    if (!isNull(expr = _parseTupleLiteral(scanner))) return expr;
    if (!isNull(expr = _parseMapLiteral(scanner))) return expr;
    return null;
  }

  Expression _parseListLiteral(SpanScanner scanner) {
    Expression literal;
    if (scanner.scan(_lBracketReg)) {
      List<Expression> list = _parseExprList(scanner);
      if (!isNull(list))
        literal = ExpressionList<dynamic>(list);
      else
        literal = const Literal<List<dynamic>>(<dynamic>[]);
      scanner.expect(_rBracketReg);
      return literal;
    }
    return null;
  }

  Expression _parseTupleLiteral(SpanScanner scanner) {
    Expression literal;
    if (scanner.scan(_lParenReg)) {
      List<Expression> list = _parseExprList(scanner);
      if (!isNull(list))
        literal = ExpressionList<dynamic>(list);
      else
        literal = const Literal<List<dynamic>>(<dynamic>[]);
      scanner.expect(_rParenReg);
      return literal;
    }
    return null;
  }

  Expression _parseMapLiteral(SpanScanner scanner) {
    if (scanner.scan(_lCurlyReg)) {
      List<String> keys = <String>[];
      List<Expression> values = <Expression>[];
      String key;
      Expression expr;
      do {
        if (scanner.scan(_stringStartReg)) {
          switch (scanner.lastMatch[1]) {
            case '"':
              scanner.expect(_stringContentDQReg,
                  name: 'Expected optional string content and double quote');
              key = scanner.lastMatch[1];
              break;
            case "'":
              scanner.expect(_stringContentSQReg,
                  name: 'Expected optional string content and single quote');
              key = scanner.lastMatch[1];
              break;
          }
          scanner.expect(_colonReg, name: 'Expected colon');
          if (isNull(expr = _parseExpr(scanner)))
            scanner.error('Expected expression');
          keys.add(key);
          values.add(expr);
        } else
          break;
      } while (scanner.scan(_commaReg));
      scanner.expect(_rCurlyReg);
      return ExpressionMap<dynamic>(
          Map<String, Expression>.fromIterables(keys, values));
    }
    return null;
  }
}

final _inReg = RegExp(_spacePlusReg.pattern + 'in' + _spacePlusReg.pattern);
// final _isReg = RegExp(_spacePlusReg.pattern + 'is' + _spacePlusReg.pattern);
// final _elseReg = RegExp(_spacePlusReg.pattern + 'else' + _spacePlusReg.pattern);
final _ifReg = RegExp(_spacePlusReg.pattern + 'if' + _spacePlusReg.pattern);
// final _concatReg = RegExp(_spaceReg.pattern + '~' + _spaceReg.pattern);
final _pipeReg = RegExp(_spaceReg.pattern + r'\|' + _spaceReg.pattern);
final _equalSignReg = RegExp(_spaceReg.pattern + '=' + _spaceReg.pattern);
final _commaReg = RegExp(_spaceReg.pattern + ',' + _spaceReg.pattern);
final _logicalOrReg =
    RegExp(_spacePlusReg.pattern + 'or' + _spacePlusReg.pattern);
final _logicalAndReg =
    RegExp(_spacePlusReg.pattern + 'and' + _spacePlusReg.pattern);
final _equalityReg =
    RegExp(_spaceReg.pattern + r'(==|!=)(?![%}])' + _spaceReg.pattern);
final _relationalReg =
    RegExp(_spaceReg.pattern + r'(>=|>|<=|<)(?![%}])' + _spaceReg.pattern);
final _additiveReg = RegExp(_spaceReg.pattern + r'(\+|-)' + _spaceReg.pattern);
final _multiplicativeReg = RegExp(
    _spaceReg.pattern + r'(\/\/|\*\*|\*|\/|%)(?![%}])' + _spaceReg.pattern);
final _unaryReg = RegExp(r'(-|\+)');
final _notReg = RegExp('(not)' + _spacePlusReg.pattern);
final _colonReg = RegExp(_spaceReg.pattern + r'\:' + _spaceReg.pattern);
final _rCurlyReg = RegExp(_spaceReg.pattern + r'\}');
final _lCurlyReg = RegExp(r'\{' + _spaceReg.pattern);
final _rBracketReg = RegExp(_spaceReg.pattern + r'\]');
final _lBracketReg = RegExp(r'\[' + _spaceReg.pattern);
final _rParenReg = RegExp(_spaceReg.pattern + r'\)');
final _lParenReg = RegExp(r'\(' + _spaceReg.pattern);
final _fieldReg = RegExp(r'\.(' + _identifierReg.pattern + ')');
final _stringContentDQReg = RegExp(r'([^"\n\r]*|\\[\n\r]*)"');
final _stringContentSQReg = RegExp(r"([^'\n\r]*|\\[\n\r]*)'");
final _stringStartReg = RegExp('(\\\'|\\\")');
final _fractionalReg = RegExp(r'(\.\d+)');
final _digitReg = RegExp(r'(\d+)');
final _identifierReg = RegExp(r'([a-zA-Z_][a-zA-Z0-9_]*)');
final _spacePlusReg = RegExp(r'[ \t]+');
final _spaceReg = RegExp(r'[ \t]*');
